package edu.kit.ipd.descartes.adaptation.model.util;

import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import org.apache.log4j.Logger;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.util.EcoreUtil;

import edu.kit.ipd.descartes.adaptation.exceptions.OperationNotPerformedException;
import edu.kit.ipd.descartes.adaptation.util.DmmModelActionHelper;
import edu.kit.ipd.descartes.core.AdaptableEntity;
import edu.kit.ipd.descartes.core.Entity;
import edu.kit.ipd.descartes.core.NamedElement;
import edu.kit.ipd.descartes.mm.adaptation.AdaptationActionOperation;
import edu.kit.ipd.descartes.mm.adaptation.AdaptationDirection;
import edu.kit.ipd.descartes.mm.adaptation.AdaptationScope;
import edu.kit.ipd.descartes.mm.adaptationpoints.ModelVariableConfigurationRange;
import edu.kit.ipd.descartes.mm.containerrepository.ContainerTemplate;
import edu.kit.ipd.descartes.mm.resourceconfiguration.ConfigurationSpecification;
import edu.kit.ipd.descartes.mm.resourceconfiguration.NumberOfParallelProcessingUnits;
import edu.kit.ipd.descartes.mm.resourcelandscape.Container;
import edu.kit.ipd.descartes.mm.resourcelandscape.RuntimeEnvironment;

public class DmmModelChanger {

    
    /**
     * Denotes a Container as a copy of another by
     * adding this mark to its name.
     */
    public static final String COPY_MARK = "-Copy";

    static Logger logger = Logger.getLogger(DmmModelChanger.class);

    /**
     * Scales the actual number of a given model variable
     * 
     * @param adaptedEntity
     * @param operation
     */
    public static void scaleModelVariable(ModelVariableConfigurationRange range, AdaptationDirection direction)
            throws OperationNotPerformedException {
        scaleModelVariable(range.getAdaptableEntity(), range, direction);
    }
    
    private static void scaleModelVariable(AdaptableEntity adaptedEntity, ModelVariableConfigurationRange range, AdaptationDirection direction)
            throws OperationNotPerformedException {
        /*
         * TODO: 1) Find all subtypes of AdaptableEntity 2) Add switch/case logic to instantiate the
         * correct class
         */
        // Here I assume that I found the correct class
        NumberOfParallelProcessingUnits adaptableModelEntity = (NumberOfParallelProcessingUnits) adaptedEntity;
        Entity containingEntity = (Entity) adaptableModelEntity.getProcessingResourceSpec().getParentResourceSpecification().eContainer();
        
        Integer currentValue = adaptableModelEntity.getNumber();
        Integer maxValue = (int) Math.floor(range.getMaxValue());
        Integer minValue = (int) Math.ceil(range.getMinValue());
        
        if (direction == AdaptationDirection.INCREASE) {
            // if current value less than max value
            if (currentValue < maxValue) {
                adaptableModelEntity.setNumber(currentValue + 1);
                logger.info("Number of " + containingEntity.getName() + " (ID: " + containingEntity.getId() + ")" + " increased to "
                        + (currentValue + 1) + ".");
            } else
                throw new OperationNotPerformedException("Adaptation operation " + direction.getLiteral()
                        + " for adaptation point " + adaptedEntity.getName()
                        + " could not be executed because the maximum has been reached");
        } else if (direction == AdaptationDirection.DECREASE) {
            // if current value greater than min value
            if (currentValue > minValue) {
                adaptableModelEntity.setNumber(currentValue - 1);
                logger.info("Number of " + containingEntity.getName() + " (ID: " + containingEntity.getId() + ")" + " decreased to "
                        + (currentValue - 1) + ".");
            } else
                throw new OperationNotPerformedException("Adaptation operation " + direction.getLiteral()
                        + " for adaptation point " + adaptedEntity.getName()
                        + " could not be executed because the minimum has been reached");
        } else
            logger.error("Adaptation direction " + direction.getLiteral() + " not defined for "
                    + adaptedEntity.eClass());
    }

    /**
     * Removes the given {@link RuntimeEnvironment} instance
     * 
     * @param runtimeEnvironment
     *            the {@link RuntimeEnvironment} to remove
     * @return
     */
    public static RuntimeEnvironment scaleInRuntimeEnvironmentInstance(RuntimeEnvironment runtimeEnvironment) {
        List<Container> similarContainers = DmmModelActionHelper.queryContainersWithSimilarTemplate(runtimeEnvironment);
        List<RuntimeEnvironment> containedEntities = DmmModelActionHelper.filterAndConvertRuntimeEnvironment(similarContainers);
        RuntimeEnvironment similarEntity = DmmModelActionHelper.filterCopiedRuntimeEnvironment(containedEntities, runtimeEnvironment);
        if (similarEntity != null) {
            runtimeEnvironment.getTemplate().getReferringContainers().remove(similarEntity);
            similarEntity.getContainedIn().getContains().remove(similarEntity);
            logger.info("RuntimeEnvironment " + similarEntity.getName() + " (ID: " + similarEntity.getId() + ") removed.");
        } else
            logger.error("Could not remove a runtime environment instance since no similar entity has been found.");

        return (RuntimeEnvironment) similarEntity;
    }

    /**
     * Creates a new {@link RuntimeEnvironment} instance
     * 
     * @param runtimeEnvironment
     *            the {@link RuntimeEnvironment} to copy
     * @return
     */
    public static RuntimeEnvironment createNewRuntimeEnvironmentInstance(RuntimeEnvironment runtimeEnvironment) {
        RuntimeEnvironment newRuntimeEnvironment = EcoreUtil.copy(runtimeEnvironment);
        newRuntimeEnvironment.setId(EcoreUtil.generateUUID());
        newRuntimeEnvironment.setName(newRuntimeEnvironment.getName() + COPY_MARK);
        newRuntimeEnvironment.setTemplate(runtimeEnvironment.getTemplate());
        runtimeEnvironment.getTemplate().getReferringContainers().add(newRuntimeEnvironment);
        newRuntimeEnvironment.getConfigSpec().clear();
        newRuntimeEnvironment.getConfigSpec().addAll(EcoreUtil.copyAll(runtimeEnvironment.getTemplate().getTemplateConfig()));
        runtimeEnvironment.getContainedIn().getContains().add(newRuntimeEnvironment);
        logger.info("New RuntimeEnvironment " + newRuntimeEnvironment.getName() + " (ID: " +
                newRuntimeEnvironment.getId() + ") to " +
                newRuntimeEnvironment.getContainedIn().getName() + " (ID: " + newRuntimeEnvironment.getContainedIn().getId() + ") added." );
        return newRuntimeEnvironment;
    }

    /**
     * Scales either the model variable if AdaptationScope is set to ALL
     * or scales a ModelEntity referring to <code>actuallyConfiguredInstance</code>
     * according to the specified <code>operation</code>.
     * 
     * @param actuallyConfiguredInstance
     * @param range
     * @param operation
     * @throws OperationNotPerformedException
     */
    public static void scaleContainerTemplate(ContainerTemplate actuallyConfiguredInstance, ModelVariableConfigurationRange range, AdaptationActionOperation operation) throws OperationNotPerformedException {

        Container container = null;
        
        AdaptationScope scope = operation.getAdaptationOperationScope();
        AdaptationDirection direction = operation.getAdaptationOperationDirection();
        
        if (operation.getAdaptationOperationScope() == AdaptationScope.ALL) {
            switch (direction) {
            case INCREASE : // falls SCALE_UP/DOWN -> ndere Template
                scaleModelVariable(range, direction);
                break;
            case DECREASE : // falls SCALE_UP/DOWN -> ndere Template
                scaleModelVariable(range, direction);
                break;
            default :
                logger.error(scope.getLiteral() + ", " + direction.getLiteral() + " not defined for " + range.getAdaptableEntity().eClass());
                break;
            }
        }
        
        else {
            switch (scope) {
            case RANDOM:
                // pick a random instance referring to this template and
                // change its configuration
                container = pickRandomReferringContainer(actuallyConfiguredInstance);
                break;
            case LEAST_UTILIZED_FIRST:
                logger.error("NOT IMPLEMENTED YET"); // TODO
                break;
            case MOST_UTILIZED_FIRST:
                logger.error("NOT IMPLEMENTED YET"); // TODO
                break;
            default:
                logger.error("No adaptation operation defined for " + scope.getLiteral() + ", "
                        + direction.getLiteral()); // TODO
                break;
            }
            if (container != null)
                try {
                    scaleResourceProvidedByContainer(container, range, direction);
                } catch (OperationNotPerformedException e) {
                    logger.error(container.getName() + " could not be adapted. Trying alternative containers.");
                    scaleResourcesOfAlternativeContainer(range, container, direction);
                }

            else
                throw new OperationNotPerformedException("No adaptable container found.");
        }
    }

    private static void scaleResourcesOfAlternativeContainer(ModelVariableConfigurationRange range, Container container,
            AdaptationDirection direction) throws OperationNotPerformedException {
        Container targetContainer = null;
        switch (direction) {
        case INCREASE:
            targetContainer = DmmModelActionHelper.queryUnfilledRuntimeEnvironment((RuntimeEnvironment) container, range.getMaxValue());
            break;
        case DECREASE:
            targetContainer = DmmModelActionHelper.queryNotEmptyRuntimeEnvironment((RuntimeEnvironment) container, range.getMinValue());
            break;
        default:
            logger.error("No corresponding AdaptationDirection found for :" + range + ".");
            break;
        }
        if (targetContainer == null) {
            logger.error("No alternative container for adaptation found!");
            throw new OperationNotPerformedException("No alternative container found. Scale operation not performed!");
        } else {
            scaleResourceProvidedByContainer(targetContainer, range, direction);
        }
    }

    
    /**Scales the resource (capacity) provided by the container, e.g., 
     * the number of virtual cores of a VM.
     * 
     * @param container The container to be scaled
     * @param range The min and max capacity
     * @param direction Specifies if up- or down-scaling
     * @throws OperationNotPerformedException
     */
    public static void scaleResourceProvidedByContainer(Container container, ModelVariableConfigurationRange range,
            AdaptationDirection direction) throws OperationNotPerformedException {

        boolean copyWasCreated = false;
        
        if (container.getConfigSpec().isEmpty()) // Better check if configSpecs are identical
        {
            copyTemplateSpecToConfigSpec(container);
            copyWasCreated = true;
        }

        NamedElement changedElement = getRespectiveAdaptableConfigSpecElement(container, range);

        if (changedElement != null && changedElement instanceof AdaptableEntity) {
            AdaptableEntity adaptableEntity = (AdaptableEntity) changedElement;
            try {
                scaleModelVariable(adaptableEntity, range, direction);
            // in case operation was not executed, e.g. because min was reached
            } catch (OperationNotPerformedException e) {
                //undo copy of config spec
                if (copyWasCreated)
                    container.getConfigSpec().clear();
                throw e;
            }
        }
        
        // If template and configSpec are identical, remove config spec
        if (EcoreUtil.equals(container.getConfigSpec(), container.getTemplate().getTemplateConfig()))
            container.getConfigSpec().clear();

    }

    private static NamedElement getRespectiveAdaptableConfigSpecElement(Container container,
            ModelVariableConfigurationRange range) {
        NamedElement changedElement = null;

        for (Iterator<ConfigurationSpecification> iterator = container.getConfigSpec().iterator(); iterator.hasNext();) {
            ConfigurationSpecification configurationSpecification = (ConfigurationSpecification) iterator.next();
            changedElement = DmmModelActionHelper.findEntityWithName(configurationSpecification, range
                    .getAdaptableEntity().getName());
            if (!changedElement.equals(null))
                break;
        }
        return changedElement;
    }

    private static void copyTemplateSpecToConfigSpec(Container container) {
        Collection<ConfigurationSpecification> copiedList = EcoreUtil.copyAll(container.getTemplate()
                .getTemplateConfig());
        container.getConfigSpec().addAll(copiedList);
    }

    private static Container pickRandomReferringContainer(ContainerTemplate template) {
        Random randomGenerator = new Random();
        EList<Container> containers = template.getReferringContainers();
        int index = randomGenerator.nextInt(containers.size());
        return containers.get(index);
    }
}
